/**
 * An abstract class that all sun time calculating classes extend. This allows
 * the algorithm used to be changed at runtime, easily allowing comparison the
 * results of using different algorithms.
 *
 * @author &copy; Eliyahu Hershfeld 2004 - 2005
 * @version 0.9.0
 */
public abstract class AstronomicalCalculator {
	private double refraction = 34 / 60d;

	// private double refraction = 34.478885263888294 / 60d;
	private double solarRadius = 16 / 60d;

	/**
	 * getDefault method returns the default sun times calculation engine.
	 *
	 * @return AstronomicalCalculator the default class for calculating sunrise
	 *         and sunset.
	 */
	public static AstronomicalCalculator getDefault() {
		return new SunTimesCalculator();
	}

	/**
	 * A method that calculates UTC sunrise as well as any time based on an
	 * angle above or below sunrise. This abstract method is implemented by the
	 * classes that extend this class.
	 *
	 * @param astronomicalCalendar
	 *            Used to calculate day of year.
	 * @param geoLocation
	 *            The GeoLocation to obtain longitude, latitude and elevation
	 *            data (elevation data might not be implemented by some
	 *            algorthms) used for the calculations.
	 * @param zenith
	 *            the azimuth below the vertical zenith of 90 degrees. for
	 *            sunrise typically the {@link #adjustZenith zenith} used for
	 *            the calculation uses geometric zenith of 90 degrees and
	 *            {@link #adjustZenith adjusts} this slightly to account for
	 *            solar refraction and the sun's radius. Another example would
	 *            be {@link AstronomicalCalendar#getBeginNauticalTwilight()}
	 *            that passes {@link AstronomicalCalendar#NAUTICAL_ZENITH} to
	 *            this method.
	 * @return The UTC time of sunrise in 24 hour format. 5:45:00 AM will return
	 *         5.75.0
	 */
	public abstract double getUTCSunrise(
			AstronomicalCalendar astronomicalCalendar, GeoLocation geoLocation,
			double zenith, boolean adjustForElevation);

	/**
	 * A method that calculates UTC sunset as well as any time based on an angle
	 * above or below sunset. This abstract method is implemented by the classes
	 * that extend this class.
	 *
	 * @param astronomicalCalendar
	 *            Used to calculate day of year.
	 * @param geoLocation
	 *            The GeoLocation to obtain longitude, latitude and elevation
	 *            data (elevation data might not be implemented by some
	 *            algorthms) used for the calculations.
	 * @param zenith
	 *            the azimuth below the vertical zenith of 90 degrees. For
	 *            sunset typically the {@link #adjustZenith zenith} used for the
	 *            calculation uses geometric zenith of 90 degrees and
	 *            {@link #adjustZenith adjusts} this slightly to account for
	 *            solar refraction and the sun's radius. Another example would
	 *            be {@link AstronomicalCalendar#getEndNauticalTwilight()} that
	 *            passes {@link AstronomicalCalendar#NAUTICAL_ZENITH} to this
	 *            method.
	 * @return The UTC time of sunset in 24 hour format. 5:45:00 AM will return
	 *         5.75.0
	 */
	public abstract double getUTCSunset(
			AstronomicalCalendar astronomicalCalendar, GeoLocation geoLocation,
			double zenith, boolean adjustForElevation);

	/**
	 * Method to return the adjustment to the zenith required to account for the
	 * elevation. Since a person at a higher elevation can see farther below the
	 * horizon, the calculation for sunrise / sunset is calculated below the
	 * horizon used at sea level. This is only used for sunrise and sunset and
	 * not times above or below it such as
	 * {@link AstronomicalCalendar#getBeginNauticalTwilight() nautical twilight}
	 * since those calculations are based on the level of available light at the
	 * given dip below the horizon, something that is not affected by elevation,
	 * the adjustment should only made if the zenith == 90
	 * {@link #adjustZenith adjusted} for refraction and solar radius.<br />
	 * The algorthm used is:
	 *
	 * <pre>
	 * elevationAdjustment = Math.toDegrees(Math.acos(earthRadiusInMeters
	 * 		/ (earthRadiusInMeters + elevationMeters)));
	 * </pre>
	 *
	 * The source of this algorthitm is <a
	 * href="http://www.calendarists.com">Calendrical Calculations</a> by
	 * Edward M. Reingold and Nachum Dershowitz. An alternate algorithm that
	 * produces an almost identical (but not accurate) result found in Ma'aglay
	 * Tzedek by Moishe Kosower and other sources is:
	 *
	 * <pre>
	 * elevationAdjustment = 0.0347 * Math.sqrt(elevationMeters);
	 * </pre>
	 *
	 * @param elevation
	 *            elevation in Meters.
	 * @return the adjusted zenith
	 */
	double getElevationAdjustment(double elevation) {
		double earthRadius = 6356.9;
		//double elevationAdjustment = 0.0347 * Math.sqrt(elevation);
		double elevationAdjustment = Math.toDegrees(Math.acos(earthRadius
				/ (earthRadius + (elevation / 1000))));
		return elevationAdjustment;
	}

	/**
	 * Adjusts the zenith to account for solar refraction, solar radius and
	 * elevation. The value for Sun's zenith and true rise/set Zenith (used in
	 * this class and subclasses) is the angle that the centre of the Sun makes
	 * to a line perpendicular to the Earth's surface. If the Sun were a point
	 * and the Earth were without an atmosphere, true sunset and sunrise would
	 * correspond to a 90-degree zenith. Because the Sun is not a point, and
	 * because the atmosphere refracts light, this 90-degree zenith does not, in
	 * fact, correspond to true sunset or sunrise, instead the centre of the
	 * Sun's disk must lie just below the horizon for the upper edge to be
	 * obscured. This means that a zenith of just above 90 degrees must be used.
	 * The Sun subtends an angle of 16 minutes of arc (this can be changed via
	 * the {@link #setSolarRadius(double)} method , and atmospheric refraction
	 * accounts for 34 minutes or so (this can be changed via the
	 * {@link #setRefraction(double)} method), giving a total of 50 arcminutes.
	 * The total value for ZENITH is 90+(5/6) or 90.8333333 degrees for true
	 * sunrise/sunset. Since a person at an elevation can see blow the horizon
	 * of a person at sea level, this will also adjust the zenith to account for
	 * elevation if available.
	 *
	 * @return The zenith adjusted to include the
	 *         {@link #getSolarRadius sun's radius},
	 *         {@link #getRefraction refraction} and
	 *         {@link #getElevationAdjustment elevation} adjustment.
	 */
	double adjustZenith(double zenith, double elevation) {
		if (zenith == AstronomicalCalendar.GEOMETRIC_ZENITH) {
			zenith = zenith
					+ (getSolarRadius() + getRefraction() + getElevationAdjustment(elevation));
		}

		return zenith;
	}

	/**
	 * Method to get the refraction value to be used when calculating sunrise
	 * and sunset. The default value is 34 arc minutes. The <a
	 * href="http://emr.cs.iit.edu/home/reingold/calendar-book/second-edition/errata.pdf">Errata
	 * and Notes for Calendrical Calculations: The Millenium Eddition</a> by
	 * Edward M. Reingold and Nachum Dershowitz lists the actual refraction
	 * value as 34.478885263888294 or approximately 34' 29". this value as well
	 * as the solarRadius and elevation adjustment are added to the zenith of
	 * sunrise and sunset.
	 *
	 * @return The refraction in arc minutes.
	 */
	double getRefraction() {
		return refraction;
	}

	/**
	 * @param refraction
	 *            The refraction in arc minutes.
	 * @see #getRefraction()
	 */
	public void setRefraction(double refraction) {
		this.refraction = refraction;
	}

	/**
	 * Method to get the sun's radius. The default value is 16 arc minutes. This
	 * will probably never be chnaged unless someone calculates that it is
	 * different than the commonly used 16.
	 *
	 * @return The sun's radius in arc minutes.
	 */
	double getSolarRadius() {
		return solarRadius;
	}

	/**
	 * Method to set the sun's radius. The default value is 16 arc minutes. This
	 * will probably never be chnaged unless someone calculates that it is
	 * different than the commonly used 16 arc minutes.
	 *
	 * @param solarRadius
	 *            The sun's radius in arc minutes.
	 */
	public void setSolarRadius(double solarRadius) {
		this.solarRadius = solarRadius;
	}
}