import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * A Java calendar that calculates astronomical time calculatons such as
 * {@link #getSunrise() sunrise} and {@link #getSunset() sunset} times. This
 * class extends GregorianCalendar and can therefore use the standard Calendar
 * functionality to change dates etc. The calculation engine used to calculate
 * the astronomical times can be changed to a different implementation by
 * implementing the {@link AstronomicalCalculator} and setting it with the
 * {@link #setAstronomicalCalculator(AstronomicalCalculator)}. A number of
 * different implementations are included in the util package <br />
 * <b>Note:</b> There are times when the algorithms can't calculate proper
 * values for sunrise and sunset. This is usually caused by trying to calculate
 * times for areas either very far North or South, where sunrise / sunset never
 * happen on that date. This is common when calculating twilight with a deep dip
 * below the horizon for locations as south as London in the northern
 * hemosphere. When the calculations encounter this condition a null will be
 * returned when a <code>Date</code> is expected and {@link Double#NaN} will
 * be returned when a double is expected. The reason that <code>Exception</code>s
 * are not thrown in these cases is due to the fact that such problems are
 * expected and will often be encountered when a calendar for a full year is
 * being generated.
 *
 * @author &copy; Eliyahu Hershfeld 2004 - 2005
 * @version 0.9.0
 */
public class AstronomicalCalendar extends GregorianCalendar {
	private static final long serialVersionUID = 1;

	/**
	 * 90 below the vertical. Used for certain calculations.<br />
	 * <b>Note </b>: it is important to note the distinction between this zenith
	 * and the regular
	 * {@link AstronomicalCalculator#adjustZenith adjusted zenith} used for most
	 * solar calculations. This 90 zenith is only used because some
	 * calculations are historically calculated as an offset in reference to 90
	 */
	public static final double GEOMETRIC_ZENITH = 90;

	/**
	 * Default value for Sun's zenith and true rise/set Zenith (used in this
	 * class and subclasses) is the angle that the centre of the Sun makes to a
	 * line perpendicular to the Earth's surface. If the Sun were a point and
	 * the Earth were without an atmosphere, true sunset and sunrise would
	 * correspond to a 90-degree zenith. Because the Sun is not a point, and
	 * because the atmosphere refracts light, this 90-degree zenith does not, in
	 * fact, correspond to true sunset or sunrise, instead the centre of the
	 * Sun's disk must lie just below the horizon for the upper edge to be
	 * obscured. This means that a zenith of just above 90 degrees must be used.
	 * The Sun subtends an angle of 16 minutes of arc (this can be changed via
	 * the {@link #setSunRadius(double)} method , and atmospheric refraction
	 * accounts for 34 minutes or so (this can be changed via the
	 * {@link #setRefraction(double)} method), giving a total of 50 arcminutes.
	 * The total value for ZENITH is 90+(5/6) or 90.8333333 degrees for true
	 * sunrise/sunset.
	 */
	// public static double ZENITH = GEOMETRIC_ZENITH + 5.0 / 6.0;
	/** Sun's zenith at civil twilight (96). */
	public static final double CIVIL_ZENITH = 96;

	/** Sun's zenith at nautical twilight (102). */
	public static final double NAUTICAL_ZENITH = 102;

	/** Sun's zenith at astronomical twilight (108). */
	public static final double ASTRONOMICAL_ZENITH = 108;

	static final long MINUTE_MILLIS = 60 * 1000;

	private GeoLocation geoLocation;

	private AstronomicalCalculator astronomicalCalculator;

	/**
	 * The getSunrise method Returns a <code>Date</code> representing the
	 * sunrise time. The zenith used for the calculation uses
	 * {@link #GEOMETRIC_ZENITH geometric zenith} of 90. This is adjusted by
	 * the {@link AstronomicalCalculator} that adds approximately 50/60 of a
	 * degree to account for 34 archminutes of refraction and 16 archminutes for
	 * the sun's radius for a total of
	 * {@link AstronomicalCalculator#adjustZenith 90.83333}. See documentation
	 * for the specific implemenatation of the {@link AstronomicalCalculator}
	 * that you are using.
	 *
	 * @return the <code>Date</code> representing the exact sunrise time. If
	 *         the calculation can not be computed null will be returned.
	 * @see AstronomicalCalculator#adjustZenith
	 */
	public Date getSunrise() {
		double sunrise = getUTCSunrise(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunrise)) {
			return null;
		} else {
			sunrise = getOffsetTime(sunrise);
			return getDateFromTime(sunrise);
		}
	}

	public Date getSeaLevelSunrise() {
		double sunrise = getUTCSeaLevelSunrise(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunrise)) {
			return null;
		} else {
			sunrise = getOffsetTime(sunrise);
			return getDateFromTime(sunrise);
		}
	}

	/**
	 * A method to return the the beginning of civil twilight (dawn) using a
	 * zenith of {@link #CIVIL_ZENITH 96}.
	 *
	 * @return The <code>Date</code> of the beginning of civil twilight using
	 *         a zenith of 96. If the calculation can not be computed null will
	 *         be returned.
	 * @see #CIVIL_ZENITH
	 */
	public Date getBeginCivilTwilight() {
		return getSunriseOffsetByDegrees(CIVIL_ZENITH);
	}

	/**
	 * A method to return the the beginning of nautical twilight using a zenith
	 * of {@link #NAUTICAL_ZENITH 102}.
	 *
	 * @return The <code>Date</code> of the beginning of nautical twilight
	 *         using a zenith of 102. If the calculation can not be computed
	 *         null will be returned.
	 * @see #NAUTICAL_ZENITH
	 */
	public Date getBeginNauticalTwilight() {
		return getSunriseOffsetByDegrees(NAUTICAL_ZENITH);
	}

	/**
	 * A method that returns the the beginning of astronomical twilight using a
	 * zenith of {@link #ASTRONOMICAL_ZENITH 108}.
	 *
	 * @return The <code>Date</code> of the beginning of astronomical twilight
	 *         using a zenith of 108. If the calculation can not be computed
	 *         null will be returned.
	 * @see #ASTRONOMICAL_ZENITH
	 */
	public Date getBeginAstronomicalTwilight() {
		return getSunriseOffsetByDegrees(ASTRONOMICAL_ZENITH);
	}

	/**
	 * The getSunset method Returns a <code>Date</code> representing the
	 * sunset time. The zenith used for the calculation uses
	 * {@link #GEOMETRIC_ZENITH geometric zenith} of 90. This is adjusted by
	 * the {@link AstronomicalCalculator} that adds approximately 50/60 of a
	 * degree to account for 34 archminutes of refraction and 16 archminutes for
	 * the sun's radius for a total of
	 * {@link AstronomicalCalculator#adjustZenith 90.83333}. See documentation
	 * for the specific implemenatation of the {@link AstronomicalCalculator}
	 * that you are using.
	 *
	 * @return the <code>Date</code> representing the exact sunset time. If
	 *         the calculation can not be computed null will be returned.
	 * @see AstronomicalCalculator#adjustZenith
	 */
	public Date getSunset() {
		double sunset = getUTCSunset(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunset)) {
			return null;
		} else {
			sunset = getOffsetTime(sunset);
			return getDateFromTime(sunset);
		}
	}

	public Date getSeaLevelSunset() {
		double sunset = getUTCSeaLevelSunset(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunset)) {
			return null;
		} else {
			sunset = getOffsetTime(sunset);
			return getDateFromTime(sunset);
		}
	}

	/**
	 * A method to return the the end of civil twilight using a zenith of
	 * {@link #CIVIL_ZENITH 96}.
	 *
	 * @return The <code>Date</code> of the end of civil twilight using a
	 *         zenith of {@link #CIVIL_ZENITH 96}. If the calculation can not
	 *         be computed null will be returned.
	 * @see #CIVIL_ZENITH
	 */
	public Date getEndCivilTwilight() {
		return getSunsetOffsetByDegrees(CIVIL_ZENITH);
	}

	/**
	 * A method to return the the end of nautical twilight using a zenith of
	 * {@link #NAUTICAL_ZENITH 102}.
	 *
	 * @return The <code>Date</code> of the end of nautical twilight using a
	 *         zenith of 102. If the calculation can not be computed null will
	 *         be returned.
	 * @see #NAUTICAL_ZENITH
	 */
	public Date getEndNauticalTwilight() {
		return getSunsetOffsetByDegrees(NAUTICAL_ZENITH);
	}

	/**
	 * A method to return the the end of astronomical twilight using a zenith of
	 * {@link #ASTRONOMICAL_ZENITH 108}.
	 *
	 * @return The The <code>Date</code> of the end of astronomical twilight
	 *         using a zenith of 108. If the calculation can not be computed
	 *         null will be returned.
	 * @see #ASTRONOMICAL_ZENITH
	 */
	public Date getEndAstronomicalTwilight() {
		return getSunsetOffsetByDegrees(ASTRONOMICAL_ZENITH);
	}

	/**
	 * Utility method that returns a time offset. This method casts the offset
	 * as a <code>long</code> and calls {@link #getTimeOffset(Date, long)}.
	 *
	 * @param time
	 *            the start time
	 * @param offset
	 *            the offset in milliseconds to add to the time
	 * @return the {@link java.util.Date}with the offset added to it
	 */
	public Date getTimeOffset(Date time, double offset) {
		return getTimeOffset(time, (long) offset);
	}

	/**
	 * A utility method to return a time offset.
	 *
	 * @param time
	 *            the start time
	 * @param offset
	 *            the offset in milliseconds to add to the time.
	 * @return the {@link java.util.Date} with the offset in milliseconds added
	 *         to it
	 */
	public Date getTimeOffset(Date time, long offset) {
		if (time == null || offset == Long.MIN_VALUE) {
			return null;
		}
		return new Date(time.getTime() + offset);
	}

	/**
	 * A utility method to return the time of an offset by degrees below or
	 * above the horizon of {@link #getSunrise() sunrise}.
	 *
	 * @param offsetZenith
	 *            the degrees before {@link #getSunrise()} to use in the
	 *            calculation. For time after sunrise use negative numbers.
	 * @return The {@link java.util.Date} of the offset after (or before)
	 *         {@link #getSunrise()}. If the calculation can not be computed
	 *         null will be returned.
	 */
	public Date getSunriseOffsetByDegrees(double offsetZenith) {
		double alos = getUTCSunrise(offsetZenith);
		if (Double.isNaN(alos)) {
			return null;
		} else {
			alos = getOffsetTime(alos);
			return getDateFromTime(alos);
		}
	}

	/**
	 * A utility method to return the time of an offset by degrees below or
	 * above the horizon of {@link #getSunset() sunset}.
	 *
	 * @param offsetZenith
	 *            the degrees after {@link #getSunset()} to use in the
	 *            calculation. For time before sunset use negative numbers.
	 * @return The {@link java.util.Date}of the offset after (or before)
	 *         {@link #getSunset()}. If the calculation can not be computed
	 *         null will be returned.
	 */
	public Date getSunsetOffsetByDegrees(double offsetZenith) {
		double tzais = getUTCSunset(offsetZenith);
		if (Double.isNaN(tzais)) {
			return null;
		} else {
			tzais = getOffsetTime(tzais);
			return getDateFromTime(tzais);
		}
	}

	/**
	 * Default constructor will set a default {@link GeoLocation} and use the
	 * default ({@link AstronomicalCalculator#getDefault()})
	 * AstronomicalCalculator.
	 */
	public AstronomicalCalendar() {
		// setGeoLocation(new GeoLocation()); //PMD
		geoLocation = new GeoLocation();
		astronomicalCalculator = AstronomicalCalculator.getDefault();
		setTimeZone(geoLocation.getTimeZone());
	}

	/**
	 * A constructor that takes in as a parameter geo location information
	 *
	 * @param geoLocation
	 *            The location information used for astronomical calculating sun
	 *            times.
	 */
	public AstronomicalCalendar(GeoLocation geoLocation) {
		this(); // call default constructor to initialize
		// astronomicalGeometryCalculator
		setGeoLocation(geoLocation);
		setTimeZone(geoLocation.getTimeZone());
	}

	/**
	 * Method that returns the sunrise in UTC time without correction for time
	 * zone offset from GMT and without using daylight savings time.
	 *
	 * @param zenith
	 *            the degrees below the horizon. For time after sunrise use
	 *            negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 */
	public double getUTCSunrise(double zenith) {
		return getAstronomicalCalculator().getUTCSunrise(this,
				getGeoLocation(), zenith, true);
	}

	public double getUTCSeaLevelSunrise(double zenith) {
		return getAstronomicalCalculator().getUTCSunrise(this,
				getGeoLocation(), zenith, false);
	}

	/**
	 * Method that returns the sunset in UTC time without correction for time
	 * zone offset from GMT and without using daylight savings time.
	 *
	 * @param zenith
	 *            the degrees below the horizon. For time after before sunset
	 *            use negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 */
	public double getUTCSunset(double zenith) {
		return getAstronomicalCalculator().getUTCSunset(this, getGeoLocation(),
				zenith, true);
	}

	public double getUTCSeaLevelSunset(double zenith) {
		return getAstronomicalCalculator().getUTCSunset(this, getGeoLocation(),
				zenith, false);
	}

	/**
	 * A method that adds time zone offset and daylight savings time to the raw
	 * UTC time.
	 *
	 * @param time
	 *            The UTC time to be adjusted.
	 * @return The time adjusted for the time zone offset and dailight savings
	 *         time.
	 */
	private double getOffsetTime(double time) {
		boolean dst = getTimeZone().inDaylightTime(getTime());
		double dstOffset = 0;
		// be nice to newfies and use a double
		double gmtOffset = getTimeZone().getRawOffset() / (60 * MINUTE_MILLIS);
		if (dst) {
			dstOffset = getTimeZone().getDSTSavings() / (60 * MINUTE_MILLIS);
		}
		return time + gmtOffset + dstOffset;
	}

	/**
	 * Method to return a temporal (solar) hour. The day from sunrise to sunset
	 * is split into 12 equal parts with each one being a temporal hour.
	 *
	 * @return the <code>long</code> millisecond length of a temporal hour. If
	 *         the calculation can not be computed {@link Long#MIN_VALUE} will
	 *         be returned.
	 */
	public long getTemporalHour() {
		return getTemporalHour(getSunrise(), getSunset());
	}

	/**
	 * Utility method that will allow the calculation of a temporal (solar) hour
	 * based on the sunrise and sunset passed to this method.
	 *
	 * @param sunrise
	 *            The start of the day.
	 * @param sunset
	 *            The end of the day.
	 * @see #getTemporalHour()
	 * @return the <code>long</code> millisecond length of the temporal hour.
	 *         If the calculation can not be computed {@link Long#MIN_VALUE}
	 *         will be returned.
	 */
	public long getTemporalHour(Date sunrise, Date sunset) {
		if (sunrise == null || sunset == null) {
			return Long.MIN_VALUE;
		}
		return (sunset.getTime() - sunrise.getTime()) / 12;
	}

	/**
	 * method that returns sundial noon. This is calculated as halfway between
	 * sunrise and sunset.
	 *
	 * @return the <code>Date</code> representing Sun's transit. If the
	 *         calculation can not be computed null will be returned.
	 */
	public Date getSunTransit() {
		// return new Date(getSunrise().getTime() + (getSolarHour() * 6));
		return getTimeOffset(getSunrise(), getTemporalHour() * 6);
	}

	/**
	 * A method that returns a <code>Date</code> from the time passed in
	 *
	 * @param time
	 *            The time to be set as the time for the <code>Date</code>.
	 *            The time expected is in the format: 18.75 for 6:45:00 PM
	 * @return The Date.
	 */
	private Date getDateFromTime(double time) {
		if (Double.isNaN(time)) {
			return null;
		}
		Calendar cal = new GregorianCalendar();
		cal.clear();
		cal.set(Calendar.YEAR, get(Calendar.YEAR));
		cal.set(Calendar.MONTH, get(Calendar.MONTH));
		cal.set(Calendar.DAY_OF_MONTH, get(Calendar.DAY_OF_MONTH));
		if ((!getTimeZone().inDaylightTime(getTime()) && time < 0) || (getTimeZone().inDaylightTime(getTime()) && time < 1)) { // sunset time (usually offset) is later than midnight GMT
			cal.add(Calendar.DAY_OF_MONTH, 1);
		}

		int hours = (int) time; // cut off minutes
		time -= hours;
		int minutes = (int) (time *= 60);
		time -= minutes;
		int seconds = (int) (time *= 60);
		time -= seconds; // milliseconds

		cal.set(Calendar.HOUR_OF_DAY, hours);
		cal.set(Calendar.MINUTE, minutes);
		cal.set(Calendar.SECOND, seconds);
		cal.set(Calendar.MILLISECOND, (int) (time * 1000));
		return cal.getTime();
	}

	public String toString() {
		return toXML();
	}

	/**
	 * @see java.lang.Object#equals(Object)
	 */
	public boolean equals(Object object) {
		if (this == object)
			return true;
		if (!(object instanceof AstronomicalCalendar))
			return false;
		AstronomicalCalendar aCal = (AstronomicalCalendar) object;
		// return Double.doubleToLongBits(ZENITH) == Double
		// .doubleToLongBits(AstronomicalCalendar.ZENITH)
		// && getGeoLocation().equals(aCal.getGeoLocation());
		return getTime().equals(aCal.getTime())
				&& getGeoLocation().equals(aCal.getGeoLocation())
				&& getAstronomicalCalculator().equals(
						aCal.getAstronomicalCalculator());
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		int result = 17;
//		 needed or this and subclasses will return identical hash
		result = 37 * result + getClass().hashCode();
		result += 37 * result + getTime().hashCode();
		result += 37 * result + getGeoLocation().hashCode();
		result += 37 * result + getAstronomicalCalculator().hashCode();
		return result;
	}

	/**
	 * A method that returns an XML formatted <code>String</code> representing
	 * the serialized <code>Object</code>. Very similar to the toString
	 * method but the return value is in an xml format. The format currently
	 * used (subject to change) is:
	 *
	 * <pre>
	 *   &lt;AstronomicalTimes date=&quot;2004-06-13&quot;&gt;
	 *       &lt;BeginAstronomicalTwilight&gt;3:23:57.479 AM&lt;/BeginAstronomicalTwilight&gt;
	 *       &lt;BeginNauticalTwilight&gt;4:12:33.200 AM&lt;/BeginNauticalTwilight&gt;
	 *       &lt;BeginCivilTwilight&gt;4:54:13.435 AM&lt;/BeginCivilTwilight&gt;
	 *       &lt;Sunrise&gt;5:27:03.640 AM&lt;/Sunrise&gt;
	 *       &lt;SunTransit&gt;12:57:06.286 PM&lt;/SunTransit&gt;
	 *       &lt;TemporalHour&gt;1.03.12345&lt;/TemporalHour&gt;
	 *       &lt;Sunset&gt;8:27:08.933 PM&lt;/Sunset&gt;
	 *       &lt;EndCivilTwilight&gt;8:59:59.978 PM&lt;/EndCivilTwilight&gt;
	 *       &lt;EndNauticalTwilight&gt;9:41:41.709 PM&lt;/EndNauticalTwilight&gt;
	 *       &lt;EndAstronomicalTwilight&gt;10:30:20.274 PM&lt;/EndAstronomicalTwilight&gt;
	 *   &lt;/AstronomicalTimes&gt;
	 * </pre>
	 *
	 * @return The XML formatted <code>String</code>.
	 */
	public String toXML() {
		StringBuffer sb = new StringBuffer();
		DateFormat df = new SimpleDateFormat("yyyy-MM-dd"); // xsd date
		TimeFormatter tf = new TimeFormatter();
		sb.append("<AstronomicalTimes date=\"").append(df.format(getSunrise()))
				.append("\">\n");
		df = new SimpleDateFormat("h:mm:ss.S a"); // time
		sb.append("\t<BeginAstronomicalTwilight>").append(
				df.format(getBeginAstronomicalTwilight())).append(
				"</BeginAstronomicalTwilight>\n");
		sb.append("\t<BeginNauticalTwilight>").append(
				df.format(getBeginNauticalTwilight())).append(
				"</BeginNauticalTwilight>\n");
		sb.append("\t<BeginCivilTwilight>").append(
				df.format(getBeginCivilTwilight())).append(
				"</BeginCivilTwilight>\n");
		sb.append("\t<Sunrise>").append(df.format(getSunrise())).append(
				"</Sunrise>\n");
		sb.append("\t<SunTransit>").append(df.format(getSunTransit())).append(
				"</SunTransit>\n");
		sb.append("\t<TemporalHour>").append(tf.format(getTemporalHour()))
				.append("</TemporalHour>\n");
		sb.append("\t<Sunset>").append(df.format(getSunset())).append(
				"</Sunset>\n");
		sb.append("\t<EndCivilTwilight>").append(
				df.format(getEndCivilTwilight())).append(
				"</EndCivilTwilight>\n");
		sb.append("\t<EndNauticalTwilight>").append(
				df.format(getEndNauticalTwilight())).append(
				"</EndNauticalTwilight>\n");
		sb.append("\t<EndAstronomicalTwilight>").append(
				df.format(getEndAstronomicalTwilight())).append(
				"</EndAstronomicalTwilight>\n");
		sb.append("</AstronomicalTimes>");
		return sb.toString();
	}

	/**
	 * @return Returns the geoLocation.
	 */
	public GeoLocation getGeoLocation() {
		return geoLocation;
	}

	/**
	 * @param geoLocation
	 *            The geoLocation to set.
	 */
	public void setGeoLocation(GeoLocation geoLocation) {
		this.geoLocation = geoLocation;
		this.setTimeZone(geoLocation.getTimeZone());// TODO might not be needed
	}

	/**
	 * A method to return the current AstronomicalCalculator set.
	 *
	 * @return Returns the astronimicalCalculator.
	 * @see #setAstronomicalCalculator(AstronomicalCalculator)
	 */
	public AstronomicalCalculator getAstronomicalCalculator() {
		return astronomicalCalculator;
	}

	/**
	 * A method to set the {@link AstronomicalCalculator} used for astronomical
	 * calculations. The Time package ships with a number of different
	 * implementations of the <code>abstract</code>
	 * {@link AstronomicalCalculator} based on different algorithms, including
	 * {@link net.sourceforge.Time.util.SunTimesCalculator one implementation}
	 * based on the <a href = "http://aa.usno.navy.mil/">US Naval Observatory's</a>
	 * algorithm, and
	 * {@link net.sourceforge.Time.util.JSuntimeCalculator another} based on
	 * <a href=""http://noaa.gov">NOAA's</a> algorithm. This allows easy
	 * runtime switching and comparison of different algorithms.
	 *
	 * @param astronomicalCalculator
	 *            The astronimicalCalculator to set.
	 */
	public void setAstronomicalCalculator(
			AstronomicalCalculator astronomicalCalculator) {
		this.astronomicalCalculator = astronomicalCalculator;
	}
}