import java.text.DecimalFormat;

// import java.text.SimpleDateFormat;

// import java.util.Date;

/**
 * A class used to format non {@link java.util.Date} times generated by the
 * Time package. For example the
 * {@link net.sourceforge.Time.AstronomicalCalendar#getTemporalHour()} returns the length of the hour
 * in milliseconds. This class can format this time.
 *
 * @author &copy; Eliyahu Hershfeld 2004 - 2005
 * @version 0.9.0
 */
public class TimeFormatter {
	private boolean prependZeroHours;

	private boolean useSeconds;

	private boolean useMillis;

	boolean useDecimal;

	private DecimalFormat minuteSecondNF;

	private DecimalFormat hourNF;

	private DecimalFormat milliNF;

	//private DecimalFormat decimalNF;

	/**
	 * Format using standard decimal format with 5 positions after the decimal.
	 */
	public static final int DECIMAL_FORMAT = 1;

	/** Format using hours and minutes. */
	public static final int SEXAGESIMAL_FORMAT = 2;

	/** Format using hours, minutes and seconds. */
	public static final int SEXAGESIMAL_SECONDS_FORMAT = 3;

	/** Format using hours, minutes, seconds and milliseconds. */
	public static final int SEXAGESIMAL_MILLIS_FORMAT = 4;

	/**
	 * Format using hours, minutes, seconds and milliseconds using the xsd:time
	 * format. This format will return 00.00.00.0 when formatting 0.
	 */
	public static final int SEXAGESIMAL_XSD_FORMAT = 5;

	public TimeFormatter() {
		this(SEXAGESIMAL_XSD_FORMAT);
	}

	/**
	 * TimeFormatter constructor using a formatter
	 *
	 * @param format
	 *            int The formatting style to use. Using
	 *            TimeFormatter.SEXAGESIMAL_SECONDS_FORMAT will format the
	 *            time time of 90*60*1000 + 1 as 1:30:00
	 */
	public TimeFormatter(int format) {
		String hourFormat = "0";
		if (prependZeroHours) {
			hourFormat = "00";
		}
		hourNF = new DecimalFormat(hourFormat);
		minuteSecondNF = new DecimalFormat("00");
		milliNF = new DecimalFormat("000");
		//decimalNF = new DecimalFormat("0.0####");
		setTimeFormat(format);
	}

	/**
	 * Sets the format to use for formatting.
	 *
	 * @param format
	 *            int the format constant to use.
	 */
	public void setTimeFormat(int format) {
		switch (format) {
		case SEXAGESIMAL_FORMAT:
			setSettings(false, false, false);
			break;
		case SEXAGESIMAL_SECONDS_FORMAT:
			setSettings(false, true, false);
			break;
		case SEXAGESIMAL_MILLIS_FORMAT:
			setSettings(false, true, true);
			break;
		case SEXAGESIMAL_XSD_FORMAT:
			setSettings(true, true, true);
			break;
		case DECIMAL_FORMAT:
		default:
			useDecimal = true;
		}
	}

	private void setSettings(boolean prependZeroHours, boolean useSeconds,
			boolean useMillis) {
		this.prependZeroHours = prependZeroHours;
		this.useSeconds = useSeconds;
		this.useMillis = useMillis;
	}

	/**
	 * A method that formats milliseconds into a time format.
	 *
	 * @param milliseconds
	 *            The time in milliseconds.
	 * @return String The formatted <code>String</code>
	 */
	public String format(double milliseconds) {
		return format((int) milliseconds);
	}

	/**
	 * A method that formats milliseconds into a time format.
	 *
	 * @param millis
	 *            The time in milliseconds.
	 * @return String The formatted <code>String</code>
	 */
	public String format(int millis) {
		return format(new Time(millis));
	}

	/**
	 * A method that formats {@link Time}objects.
	 *
	 * @param time
	 *            The time <code>Object</code> to be formatted.
	 * @return String The formatted <code>String</code>
	 */
	public String format(Time time) {
		StringBuffer sb = new StringBuffer();
		sb.append(hourNF.format(time.getHours()));
		sb.append(":");
		sb.append(minuteSecondNF.format(time.getMinutes()));
		if (useSeconds) {
			sb.append(":");
			sb.append(minuteSecondNF.format(time.getSeconds()));
		}
		if (useMillis) {
			sb.append(".");
			sb.append(milliNF.format(time.getMilliseconds()));
		}
		return sb.toString();
	}

	/*
	 * public String format(Date date) { if (date == null) { return "N/A"; }
	 * else { return dateFormat.format(date); } }
	 */

}